library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
library project_lib;
use project_lib.histpack.all;

entity histRAMController is
  port (clkIn             : in std_logic;
	clk2xIn           : in std_logic;
	resetLowIn        : in std_logic;
	vertBlankIn       : in std_logic;
	pixelIn           : in PixelDataType;  -- video RAM data bus
	vidCalculateIn    : in Boolean; -- video RAM in CALCULATE mode
	vidReadIn         : in Boolean;
	histAddrOut       : out PixelDataType;
	histWriteEnLowOut : out std_logic;
	histChipSelLowOut : out std_logic;
	histClearOut      : out boolean; -- histRAM is still being cleared
	histReadoutOut    : out boolean; -- histogram is being sequenced
	histDataBi        : inout PixelCountType
	);

  type HistState is (IDLE, CLEAR, READWAIT, CALCULATE, READOUT);
end histRAMController;

architecture rtl of histRAMController is
  signal addressRSig  : PixelDataType;
  signal nextAddrSig  : PixelDataType;
  signal stateRSig    : HistState;
  signal nextStateSig : HistState;
  signal nextBin      : PixelCountType;

  signal writeEnable  : boolean;

  signal halfClkSig : std_logic;
begin
  
  addGen: process (stateRSig, vertBlankIn, pixelIn, 
		   clkIn, clk2xIn, halfClkSig, vidReadIn)
  begin
    -- assert PixelDataRange'high >= PixelCountRange'high
    -- (or the histogram RAM will be cleared before the image is read in)

    histClearOut <= FALSE;
    histReadoutOut <= FALSE;
    case stateRSig is
      when IDLE =>    -- wait for end of vertical blank
	histChipSelLowOut <= '1';
	histWriteEnLowOut <= '1';
	nextAddrSig <= (others => '0');
	histDataBi <= (others => 'Z');
	if (vertBlankIn = '0') then
	  nextStateSig <= CLEAR;
	else
	  nextStateSig <= IDLE;
	end if;
      when CLEAR =>
	histChipSelLowOut <= not (clkIn xor clk2xIn);
	histWriteEnLowOut <= '0';
	histDataBi <= (others => '0');
	nextAddrSig <= addressRSig + "1";  -- next address
	histClearOut <= TRUE;
	if(addressRSig = PixelDataRange'high) then -- last memory address
	  if(vidReadIn = FALSE) then
	    nextStateSig <= CALCULATE;
	    nextAddrSig <= pixelIn;
	    histClearOut <= FALSE;
	    nextAddrSig <= pixelIn;
	  else
	    nextStateSig <= READWAIT; -- wait for image to finish
	    nextAddrSig <= (others => 'Z');
	  end if;
	else
	  nextStateSig <= CLEAR;
	end if;
      when READWAIT => -- wait for videoRAMctl to finish storing image
	histChipSelLowOut <= '1';
	histWriteEnLowOut <= '1';
	histDataBi <= (others => 'Z');
	nextAddrSig <= (others => 'Z');
	if (vidReadIn = FALSE) then
	  nextStateSig <= CALCULATE;
	  nextAddrSig <= (others => '0'); -- must be something, but it doesn't 
                                          -- matter what (all are value 0)
	else
	  nextStateSig <= READWAIT;
	end if;
      when CALCULATE =>
	histChipSelLowOut <= not (clkIn xor clk2xIn);
	histWriteEnLowOut <= halfClkSig;
	nextAddrSig <= pixelIn;
	if (halfClkSig = '1') then 	
	  histDataBi <= (others => 'Z');
	else
	  histDataBi <= nextBin;
	end if;
	if(vidCalculateIn = TRUE or halfClkSig = '1') then
	  nextStateSig <= CALCULATE;
	else
	  nextAddrSig <= (others => '0'); -- prepare for readout
	  nextStateSig <= READOUT;
	end if;
      when READOUT =>
	histChipSelLowOut <= '0'; -- not (clkIn xor clk2xIn);
	histWriteEnLowOut <= '1';
	histDataBi <= (others => 'Z');
	histReadoutOut <= TRUE;
	nextAddrSig <= addressRSig + "1"; -- next address
	if (addressRSig = PixelDataRange'high) then
	  nextStateSig <= IDLE;
	else
	  nextStateSig <= READOUT;
	end if;
    end case;
    
  end process; -- addGen
  
  incrementBin: process(clkIn, clk2xIn, stateRSig, resetLowIn)
  begin
    if (resetLowIn = '0') then
      nextBin <= (others => '0');
    elsif (stateRSig = CALCULATE and clk2xIn'event and 
	   clk2xIn = '0' and clkIn = '0') then
      nextBin <= histDataBi + "1";
    end if;
  end process; -- incrementBin

  latchState: process (clkIn, resetLowIn)
  begin
    if(resetLowIn = '0') then
      stateRSig <= IDLE;
      addresSRSig <= (others => '0');
      halfClkSig <= '1';
    elsif(clkIn'event and clkIn = '1') then
      stateRSig <= nextStateSig;
      addressRSig <= nextAddrSig;
      halfClkSig <= not halfClkSig;
    end if;
  end process;

  histAddrOut <= addressRSig;

end rtl; -- histRAMController
