library ieee;
use ieee.std_logic_1164.all;
library project_lib;
use project_lib.histpack.all;

entity histogram is
  port(
       clk2xIn            : in std_logic;
       clkIn              : in std_logic;
       cpuAddrIn          : in CpuAddrType;
       cpuDataStrobeLowIn : in std_logic;
       cpuWriteEnLowIn    : in std_logic;
       horizBlankIn       : in std_logic;
       pixelDataIn        : in PixelDataType;
       resetLowIn         : in std_logic;
       vertBlankIn        : in std_logic;
       eofIntLowOut       : out std_logic;
       histAddrOut        : out PixelDataType;
       histChipSelLowOut  : out std_logic;
       histWriteEnLowOut  : out std_logic;
       statDataOut        : out PixelCountType;
       statDataValidOut   : out std_logic;
       vidAddrOut         : out PixelCountType;
       vidChipSelLowOut   : out std_logic;
       vidWriteEnLowOut   : out std_logic;
       cpuDataBi          : inout CpuDataType;
       histDataBi         : inout PixelCountType;
       vidDataBi          : inout PixelDataType
       );

  type EofInterruptState is (CLEAR, ASSERTED, HALFCLEAR);
end histogram;

architecture structure of histogram is
  signal histClearSig   : Boolean; 
  signal histReadoutSig : Boolean; 
  signal histComputeSig : Boolean;
  signal vidReadSig     : Boolean;

  signal pauseRSig      : Boolean;
  signal nextPauseSig   : Boolean;
  
  signal thresholdRSig   : PixelDataType;
  signal nextThresholdSig : PixelDataType;
  signal thresholdPixelDataSig: PixelDataType;

  signal eofAckState      : EofInterruptState := ASSERTED;
  signal nextEofAckState  : EofInterruptState;
  signal assertEofSig     : Boolean;
  signal statDataValidSig : Boolean;

  signal maskedFrameRSig    : Boolean;
  signal maskedVertBlankSig : std_logic;

  -- the following signals are driven by the histRAMctl and vidRAMctl.
  -- They are only propagated to the output by this module when the CPU is
  -- not driving them.
  signal vidAddrSig        : PixelCountType;
  signal vidChipSelLowSig  : std_logic;
  signal vidWriteEnLowSig  : std_logic;
  signal histAddrSig       : PixelDataType;
  signal histChipSelLowSig : std_logic;
  signal histWriteEnLowSig : std_logic;

  component videoRAMController
    port (clkIn            : in std_logic;
	  clk2xIn          : in std_logic;
	  resetLowIn       : in std_logic;
	  histClearIn      : in Boolean;
	  horizBlankIn     : in std_logic;
	  vertBlankIn      : in std_logic;
	  pixelDataIn      : in PixelDataType;
	  vidAddrOut       : out PixelCountType;
	  vidWriteEnLowOut : out std_logic;
	  vidChipSelLowOut : out std_logic;
	  histComputeOut   : out Boolean;
	  vidReadOut       : out Boolean;
	  vidDataBi        : inout PixelDataType
	  );
  end component;
  
  for videoRAMctl: videoRAMController
    use entity WORK.videoRAMController(rtl);

  component histRAMController
    port (clkIn             : in std_logic;
	  clk2xIn           : in std_logic;
	  resetLowIn        : in std_logic;
	  vertBlankIn       : in std_logic;
	  pixelIn           : in PixelDataType;
	  vidCalculateIn    : in Boolean;
	  vidReadIn         : in Boolean;
	  histAddrOut       : out PixelDataType;
	  histWriteEnLowOut : out std_logic;
	  histChipSelLowOut : out std_logic;
	  histClearOut      : out Boolean;
	  histReadoutOut    : out Boolean;
	  histDataBi        : inout PixelCountType
	  );
    end component;

  for histRAMctl: histRAMController
    use entity WORK.histRAMController(rtl);

  component medianController
    port (clkIn            : in std_logic;
	  clk2xIn          : in std_logic;
	  resetLowIn       : in std_logic;
	  histDataIn       : in PixelCountType;
	  histAddrIn       : in PixelDataType;
	  histComputeIn    : in Boolean;
	  histReadoutIn    : in Boolean;
	  statDataOut      : out PixelCountType;
	  statDataValidOut : out Boolean;
	  assertEofOut     : out Boolean
	  );
  end component;
  
  for medianCtl: medianController
    use entity WORK.medianController(rtl);

begin
  videoRAMctl: videoRAMController
    port map(clkIn             => clkIn,
	     clk2xIn           => clk2xIn,
	     resetLowIn        => resetLowIn,
	     histClearIn       => histClearSig,
	     horizBlankIn      => horizBlankIn,
	     vertBlankIn       => maskedVertBlankSig,
	     pixelDataIn       => pixelDataIn,
	     vidAddrOut        => vidAddrSig,
	     vidWriteEnLowOut  => vidWriteEnLowSig,
	     vidChipSelLowOut  => vidChipSelLowSig,
	     histComputeOut    => histComputeSig,
	     vidReadOut        => vidReadSig,
	     vidDataBi         => vidDataBi
	     );

  histRAMctl: histRAMController
    port map(clkIn             => clkIn,
	     clk2xIn           => clk2xIn,
	     resetLowIn        => resetLowIn,
	     vertBlankIn       => maskedVertBlankSig,
	     pixelIn           => thresholdPixelDataSig,
	     vidCalculateIn    => histComputeSig,
	     vidReadIn         => vidReadSig,
	     histAddrOut       => histAddrSig,
	     histWriteEnLowOut => histWriteEnLowSig,
	     histChipSelLowOut => histChipSelLowSig,
	     histClearOut      => histClearSig,
	     histReadoutOut    => histReadoutSig,
	     histDataBi        => histDataBi
	     );

  medianCtl: medianController
    port map(clkIn             => clkIn,
	     clk2xIn           => clk2xIn,
	     resetLowIn        => resetLowIn,
	     histDataIn        => histDataBi,
	     histAddrIn        => histAddrSig,
	     histComputeIn     => histComputeSig,
	     histReadoutIn     => histReadoutSig,
	     statDataOut       => statDataOut,
	     statDataValidOut  => statDataValidSig,
	     assertEofOut      => assertEofSig
	     );

   drivers: process(statDataValidSig)
   begin
     if (statDataValidSig = TRUE) then
       statDataValidOut <= '1';
     else
       statDataValidOut <= '0';
     end if;
   end process; -- drivers

  -- disable frame processing by masking vertical blank
  latchMaskFrame: process (vertBlankIn)
  begin
    if(vertBlankIn'event and vertBlankIn = '0') then
      maskedFrameRSig <= pauseRSig;
    end if;
  end process; -- maskVertical

  maskVert: process (vertBlankIn, maskedFrameRSig)
  begin
    if (maskedFrameRSig = TRUE) then
      maskedVertBlankSig <= '1';
    else
      maskedVertBlankSig <= vertBlankIn;
    end if;
  end process; -- maskVert

  -- provide access to RAM busses in pause mode
  cpuRamAccess : process (cpuAddrIn, cpuWriteEnLowIn, cpuDataBi, 
			  clkIn, clk2xIn, vidChipSelLowSig, vidWriteEnLowSig,
			  histChipSelLowSig, histWriteEnLowSig, 
			  histAddrSig, vidAddrSig)
    variable addressBase : CpuAddrType;
  begin
    addressBase := cpuAddrIn;
    addressBase(15 downto 0) := (others => '0'); -- mask low order stuff
    histDataBi <= (others => 'Z');
    histAddrOut <= histAddrSig;
    histChipSelLowOut <= histChipSelLowSig;
    histWriteEnLowOut <= histWriteEnLowSig;
    vidDataBi <= (others => 'Z');
    vidAddrOut <= vidAddrSig;
    vidChipSelLowOut <= vidChipSelLowSig;
    vidWriteEnLowOut <= vidWriteEnLowSig;
    cpuDataBi <= (others => 'Z');
    if (maskedFrameRSig = TRUE and cpuWriteEnLowIn = '0') then
      case addressBase is
	when VIDRAMBASE =>
	  vidAddrOut <= cpuAddrIn(PixelCountBvRange);
	  vidDataBi <= cpuDataBi(PixelDataBvRange);
	  vidWriteEnLowOut <= '0';
	  vidChipSelLowOut <= cpuDataStrobeLowIn;
	when HISTRAMBASE =>
	  histAddrOut <= cpuAddrIn(PixelDataBvRange);
	  histDataBi <= cpuDataBi(PixelCountBvRange);
	  histWriteEnLowOut <= '0';
	  histChipSelLowOut <= cpuDataStrobeLowIn;
	when others =>
      end case;
    elsif (maskedFrameRSig = TRUE and cpuWriteEnLowIn = '1') then
      case addressBase is
	when VIDRAMBASE =>
	  vidAddrOut <= cpuAddrIn(PixelCountBvRange);
	  vidChipSelLowOut <= '0';
	  cpuDataBi <= (others => '0');
	  cpuDataBi(PixelDataBvRange) <= vidDataBi;
	when HISTRAMBASE =>
	  histAddrOut <= cpuAddrIn(PixelDataBvRange);
	  histChipSelLowOut <= '0';
	  cpuDataBi <= (others => '0');
	  cpuDataBi(PixelCountBvRange) <= histDataBi;
	when others =>
      end case;
    end if;
  end process; -- cpuRamAccess

  -- deal with threshold register reading and writing
  thresholdStuff: process (cpuAddrIn, cpuWriteEnLowIn, cpuDataBi)
  begin
    if (cpuAddrIn = THRESHREG and cpuWriteEnLowIn = '0') then
      nextThresholdSig <= cpuDataBi(PixelDataBvRange);
      cpuDataBi <= (others => 'Z');
    elsif (cpuAddrIn = THRESHREG and cpuWriteEnLowIn = '1') then
      cpuDataBi <= (others => '0');
      cpuDataBi(PixelDataBvRange) <= thresholdRSig;
      nextThresholdSig <= thresholdRSig;
    else
      nextThresholdSig <= thresholdRSig;
      cpuDataBi <= (others => 'Z');
    end if;
  end process; -- thresholdStuff

  -- deal with actually thresholding the data
  thresholdPixel: process(vidDataBi)
  begin
    if (vidDataBi > thresholdRSig) then
      thresholdPixelDataSig <= thresholdRSig;
    else
      thresholdPixelDataSig <= vidDataBi;
    end if;
  end process; -- thresholdPixel

  -- deal with pause register reading and writing
   pauseStuff: process (cpuAddrIn, cpuWriteEnLowIn, cpuDataBi)
   begin
     if (cpuAddrIn = PAUSEFLAG and cpuWriteEnLowIn = '0') then
       if(cpuDataBi(0) = '0') then
	 nextPauseSig <= FALSE;
       else
	 nextPauseSig <= TRUE;
       end if;
       cpuDataBi <= (others => 'Z');
     elsif (cpuAddrIn = PAUSEFLAG and cpuWriteEnLowIn = '1') then
       cpuDataBi <= (others => '0');
       if (pauseRSig = TRUE) then
	 cpuDataBi(0) <= '1';
       else
	 cpuDataBi(0) <= '0';
       end if;
     else
       nextPauseSig <= pauseRSig;
       cpuDataBi <= (others => 'Z');
     end if;
   end process; -- pauseStuff
  
  -- deal with interrupt acknowledge from CPU
  eofStuff: process (eofAckState, cpuWriteEnLowIn, assertEofSig)
  begin
    case eofAckState is
      when CLEAR =>
	eofIntLowOut <= '1';
	nextEofAckState <= CLEAR;
      when ASSERTED =>
	eofIntLowOut <= '0';
	if (cpuAddrIn = EOFINTACK and cpuWriteEnLowIn = '0') then
	  nextEofAckState <= HALFCLEAR;
	else
	  nextEofAckState <= ASSERTED;
	end if;
      when HALFCLEAR =>
	eofIntLowOut <= '0';
	if(cpuAddrIn = EOFINTACK and cpuWriteEnLowIn = '0' 
	   and cpuDataBi(0) = '1') then
	  nextEofAckState <= CLEAR;
	elsif(cpuAddrIn = EOFINTACK and cpuWriteEnLowIn = '0' and 
	      cpuDataBi(0) /= '1') then
	  nextEofAckState <= ASSERTED;
	else
	  nextEofAckState <= HALFCLEAR;
	end if;
    end case;
  end process; -- eofStuff

  latchStuff: process (cpuDataStrobeLowIn)
  begin
    if(cpuDataStrobeLowIn'event and 
       (cpuDataStrobeLowIn = '1' or cpuDataStrobeLowIn = 'H')) then
      thresholdRSig <= nextThresholdSig;
      pauseRSig <= nextPauseSig;
    end if;
  end process; -- latchStuff

  latchEofState: process(cpuDataStrobeLowIn, assertEofSig)
  begin
    if (assertEofSig'event and assertEofSig = FALSE) then
      eofAckState <= ASSERTED;
    elsif (cpuDataStrobeLowIn'event 
	and (cpuDataStrobeLowIn = '1' or cpuDataStrobeLowIn = 'H')) then
      eofAckState <= nextEofAckState;
    end if;
  end process; -- latchEofState
end structure; -- histogram architecture


