library ieee;
use ieee.std_logic_1164.all;
package histpack is
   subtype CpuAddrBvRange is integer range 23 downto 0;
   subtype CpuAddrType is std_logic_vector(CpuAddrBvRange);

   subtype CpuDataBvRange is integer range 15 downto 0;
   subtype CpuDataType is std_logic_vector(CpuDataBvRange);

   -- Max size of the image in "X" direction
   subtype PixelXRange is integer range 15 downto 0;
   subtype PixelXBvRange is integer range 3 downto 0;

   -- Max size of the image in "Y" direction
   subtype PixelYRange is integer range 15 downto 0;
   subtype PixelYBvRange is integer range 3 downto 0;

   -- Types to describe the pixel data
   subtype PixelDataRange is integer range 255 downto 0;
   subtype PixelDataBvRange is integer range 7 downto 0;
   subtype PixelDataType is std_logic_vector(PixelDataBvRange);

   -- Type for Video and Histogram RAMs
   subtype PixelCountRange is integer range
         ((PixelXRange'high + 1) * (PixelYRange'high + 1)) downto 0;
   subtype PixelCountBvRange is integer range
         (PixelXBvRange'high + PixelYBvRange'high + 1) downto 0;
   subtype PixelCountType is std_logic_vector(PixelCountBvRange);

   -- Only the top 3 address bits are used to decode memory blocks
   constant THRESHREG   : CpuAddrType := "100000000000000000000000"; -- 80,0000h
                                          ----    ----    ----    
   constant PAUSEFLAG   : CpuAddrType := "100000000000000000000001"; -- 80,0001h
                                          ----    ----    ----    
   constant EOFINTACK   : CpuAddrType := "100000000000000000000010"; -- 80,0002h
                                          ----    ----    ----    
   constant VIDRAMBASE  : CpuAddrType := "101000000000000000000000"; -- A0,0000h
                                          ----    ----    ----    
   constant HISTRAMBASE : CpuAddrType := "110000000000000000000000"; -- C0,0000h
                                          ----    ----    ----    
   procedure CpuWrite(constant addrToWrite : in  CpuAddrType;
                      constant dataToWrite : in  CpuDataType;
                      signal   cpuAddr     : out CpuAddrType;
                      signal   cpuData     : out CpuDataType;
                      signal   cpuWeLow    : out std_logic;
                      signal   cpuCsLow    : out std_logic
                     );

   procedure CpuRead(constant addrToWrite : in  CpuAddrType;
                     variable dataToRead  : out CpuDataType;
                     signal   cpuAddr     : out CpuAddrType;
                     signal   cpuData     : in  CpuDataType;
                     signal   cpuWeLow    : out std_logic;
                     signal   cpuCsLow    : out std_logic
                     );

   type HistArray is array(PixelDataRange) of integer;

   type SimDataRecord is record
      x      : PixelXRange;
      y      : PixelYRange;
      thresh : PixelDataRange;
   end record;

   type SimDataArray is array (3 downto 0) of SimDataRecord;

   constant SIMRUNDATA : SimDataArray :=
         ( 0 => (x =>  4, y =>  3, thresh => 214),
           1 => (x => 15, y => 15, thresh =>  50),
           2 => (x =>  8, y => 11, thresh => 255),
           3 => (x =>  9, y => 13, thresh => 188) );

end histpack;

package body histpack is

   procedure CpuWrite(constant addrToWrite : in  CpuAddrType;
                      constant dataToWrite : in  CpuDataType;
                      signal   cpuAddr     : out CpuAddrType;
                      signal   cpuData     : out CpuDataType;
                      signal   cpuWeLow    : out std_logic;
                      signal   cpuCsLow    : out std_logic
                     ) is
   begin
      cpuAddr  <= addrToWrite after 3 ns;
      cpuData  <= dataToWrite after 3 ns;
      cpuWeLow <= '0' after 5 ns;
      wait for 100 ns;
      cpuCsLow <= '0' after 3 ns;
      wait for 100 ns;
      cpuCsLow <= '1' after 2 ns;
      wait for 100 ns;
      cpuAddr  <= (others => 'Z') after 3 ns;
      cpuData  <= (others => 'Z') after 3 ns;
      cpuWeLow <= '1' after 5 ns;
      wait for 100 ns;
   end CpuWrite;
   
   procedure CpuRead(constant addrToWrite : in  CpuAddrType;
                     variable dataToRead  : out CpuDataType;
                     signal   cpuAddr     : out CpuAddrType;
                     signal   cpuData     : in  CpuDataType;
                     signal   cpuWeLow    : out std_logic;
                     signal   cpuCsLow    : out std_logic
                     ) is
   begin
      cpuAddr  <= addrToWrite after 3 ns;
      cpuWeLow <= '0' after 5 ns;
      wait for 100 ns;
      cpuCsLow <= '0' after 3 ns;
      wait for 100 ns;
      dataToRead := cpuData;
      cpuCsLow <= '1' after 2 ns;
      wait for 100 ns;
      cpuAddr  <= (others => 'Z') after 3 ns;
      cpuWeLow <= '1' after 5 ns;
      wait for 100 ns;
   end CpuRead;

end histpack;
