library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
library project_lib;
use project_lib.histpack.all;

entity medianController is
  port (clkIn          : in std_logic;
	clk2xIn        : in std_logic;
	resetLowIn     : in std_logic;
	histDataIn     : in PixelCountType;
	histAddrIn     : in PixelDataType;  -- histogram RAM address bus
	histComputeIn  : in Boolean; -- TRUE when histogram is being computed
	histReadoutIn  : in Boolean; -- TRUE when histogram is being sequenced
	statDataOut      : out PixelCountType;
	statDataValidOut : out Boolean;
	assertEofOut     : out Boolean
	);

  type StatisticState is (IDLE, COUNT, SEQUENCE, MEDIAN);
end medianController;

architecture rtl of medianController is
  signal accumulatorRSig    : Integer;
  signal nextAccumulatorSig : Integer;
  signal medianRSig         : PixelCountType;
  signal assertMedianRSig   : Boolean;
  signal medianDetectedSig  : Boolean;
  
  signal nextHalfImageSizeSig : Integer;
  signal halfImageSizeRSig   : Integer;
  signal clockDividerRSig    : Integer;
  
  signal stateRSig           : StatisticState;
  signal nextStateSig        : StatisticState;
begin

  statistics: process(stateRSig, histReadoutIn, histDataIn, 
		      clockDividerRSig, halfImageSizeRSig, medianRSig)
  begin
    case stateRSig is
      when IDLE =>
	statDataOut <= (others => 'Z');
	statDataValidOut <= FALSE;
	nextHalfImageSizeSig <= 0;
	assertEofOut <= FALSE;
	if (histComputeIn) then
	  nextStateSig <= COUNT;
	else
	  nextStateSig <= IDLE;
	end if;
      when COUNT =>
	statDataOut <= (others => 'Z');
	statDataValidOut <= FALSE;
	if (clockDividerRSig = 3) then
	  nextHalfImageSizeSig <= halfImageSizeRSig + 1;
	else
	  nextHalfImageSizeSig <= halfImageSizeRSig;
	end if;
	assertEofOut <= FALSE;
	if (histReadoutIn) then
	  nextStateSig <= SEQUENCE;
	  statDataOut <= histDataIn;
	  statDataValidOut <= TRUE;
	else
	  nextStateSig <= COUNT;
	end if;
      when SEQUENCE => -- write out histogram
	statDataOut <= histDataIn;
	assertEofOut <= FALSE;
	nextHalfImageSizeSig <= halfImageSizeRsig;
	if (not histReadoutIn) then
	  nextStateSig <= MEDIAN;
	  statDataValidOut <= FALSE;
	else
	  nextStateSig <= SEQUENCE;
	  statDataValidOut <= TRUE;
	end if;
      when MEDIAN =>   -- write out median value
	statDataOut <= medianRSig;
	statDataValidOut <= FALSE;
	nextHalfImageSizeSig <= halfImageSizeRsig;
	assertEofOut <= TRUE;
	nextStateSig <= IDLE;
    end case;
  end process; -- statistics

  latchState: process(clkIn, resetLowIn)
  begin
    if (resetLowIn = '0') then
      stateRSig <= IDLE;
      halfImageSizeRSig <= 0;
      clockDividerRSig <= 0;
    elsif (clkIn'event and clkIn = '1') then
      stateRSig <= nextStateSig;
      halfImageSizeRSig <= nextHalfImageSizeSig;
      if (clockDividerRSig = 3) then
	clockDividerRSig <= 0;
      else
	clockDividerRSig <= clockDividerRSig + 1;
      end if;
    end if;
  end process; -- latchState

  -- clock in in the next accumulator value
  latchAccumulator: process (clk2xIn, resetLowIn, stateRSig)
  begin
    if (resetLowIn = '0' or stateRSig = IDLE) then
      accumulatorRSig <= 0;
    elsif(clk2xIn'event and clk2xIn = '0' and clkIn = '0' and 
	  nextStateSig = SEQUENCE) then
      accumulatorRSig <= accumulatorRSig + to_integer(histDataIn);
    end if;
  end process; -- latchAccumulator

  medianDetectedSig <= accumulatorRSig >= halfImageSizeRSig;
  detectMedian: process(medianDetectedSig)
  begin
    if (medianDetectedSig'event and medianDetectedSig = TRUE) then
      medianRSig <= histAddrIn;
    end if;

  end process; -- detectMedian

end rtl; -- medianController
