library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
library project_lib;
use project_lib.histpack.all;

entity videoRAMController is
  port (clkIn            : in std_logic;
	clk2xIn          : in std_logic;
	resetLowIn       : in std_logic;
	histClearIn      : in Boolean;
	horizBlankIn     : in std_logic;
	vertBlankIn      : in std_logic;
	pixelDataIn      : in PixelDataType;
	vidAddrOut       : out PixelCountType;
	vidWriteEnLowOut : out std_logic;
	vidChipSelLowOut : out std_logic;
	histComputeOut   : out Boolean;
	vidReadOut       : out Boolean;
	vidDataBi        : inout PixelDataType
	);

  type VideoState is (IDLE, READIN, HORIZBLANK, CLEARWAIT, CALCULATE);
end videoRAMController;

architecture rtl of videoRAMController is
  signal addressRSig     : PixelCountType;
  signal nextAddrSig     : PixelCountType;
  signal stateRSig       : VideoState;
  signal nextStateSig    : VideoState;

  signal imageXSizeRSig   : std_logic_vector(PixelXBvRange);
  signal imageYSizeRSig   : std_logic_vector(PixelYBvRange);

  signal writeEnable     : Boolean;
  
  signal halfClkSig      : std_logic;
begin

  -- implement the address generator as a counter that increments with
  -- the rising clkIn edge when neither blank pulse is asserted.  As
  -- the address in incremented, the current pixel value is
  -- simultaneously written to video RAM.  The counter is cleared
  -- during every vertical blank.
  addGen: process(stateRSig, horizBlankIn, vertBlankIn, pixelDataIn, clk2xIn)
  begin
    writeEnable <= FALSE;
    vidChipSelLowOut <= '1';
    vidDataBi <= (others => 'Z');
    histComputeOut <= FALSE;
    vidReadOut <= FALSE;
    case stateRSig is
      when IDLE =>
	if(vertBlankIn = '1') then
	  nextAddrSig <= (others => '0');
	  nextStateSig <= IDLE;
	else
	  writeEnable <= TRUE;
	  vidChipSelLowOut <= not (clkIn xor clk2xIn);
	  vidDataBi <= pixelDataIn;
	  nextAddrSig <= (0 => '1', others => '0');
	  nextStateSig <= READIN;
	end if;
      when READIN =>
	writeEnable <= TRUE;
	vidReadOut <= TRUE;
	-- the address counter jumps to the next line at horiz refresh
	if (vertBlankIn = '1') then
	  nextAddrSig <= (others => '0');
	  if (histClearIn = TRUE) then
	    nextStateSig <= CLEARWAIT;
	  else
	    vidReadOut <= FALSE;
	    writeEnable <= FALSE;
	    nextStateSig <= CALCULATE;
	  end if;
	elsif (horizBlankIn = '1') then
	  if to_integer(addressRSig(PixelXBvRange'high downto 0)) /= 0 then
	    nextAddrSig(PixelXBvRange'high downto 0) <= (others => '0'); 
	    nextAddrSig(PixelCountType'high downto (PixelXBvRange'high + 1)) <=
	      addressRSig(PixelCountType'high downto 
			  (PixelXBvRange'high + 1)) + "1"; -- next line
	  else
	    nextAddrSig <= addressRSig;
	  end if;
	  nextStateSig <= HORIZBLANK;
	else
	  vidChipSelLowOut <= not (clkIn xor clk2xIn);
	  nextAddrSig <= addressRSig + "1"; -- next pixel
	  vidDataBi <= pixelDataIn;
	  nextStateSig <= READIN;
	end if;
      when HORIZBLANK =>
	writeEnable <= TRUE;
	vidReadOut <= TRUE;
	if(horizBlankIn = '0') then
	  nextStateSig <= READIN;
	  vidChipSelLowOut <= not (clkIn xor clk2xIn);
	  nextAddrSig <= addressRSig + "1";
	  vidDataBi <= pixelDataIn;
	else
	  nextStateSig <= HORIZBLANK;
	  nextAddrSig <= addressRSig;
	end if;
      when CLEARWAIT =>
	nextAddrSig <= (others => '0');
	if(histClearIn = TRUE) then
	  nextStateSig <= CLEARWAIT;
	else
	  nextStateSig <= CALCULATE;
	end if;
      when CALCULATE =>
	vidChipSelLowOut <= '0';
	histComputeOut <= TRUE;
	if (halfClkSig = '0') then
	  if(addressRSig(PixelXBvRange'high downto 0) = imageXSizeRSig) then
	    nextAddrSig(PixelXBvRange'high downto 0) <= (others => '0');
	    nextAddrSig(PixelCountType'high downto (PixelXBvRange'high + 1)) <=
	      addressRSig(PixelCountType'high downto 
			  (PixelXBvRange'high + 1)) + "1"; -- next line
	  else
	    nextAddrSig <= addressRSig + "1";
	  end if;
	else
	  nextAddrSig <= addressRSig;
	end if;
	if(addressRSig = imageYSizeRSig & imageXSizeRSig and 
	   halfClkSig = '0') then
	  nextStateSig <= IDLE;
	else
	  nextStateSig <= CALCULATE;
	end if;
    end case;
  end process; -- addGen


  latchImageSize: process(clkIn, resetLowIn, horizBlankIn)
  begin
    if(resetLowIn = '0') then
      imageXSizeRSig <= (others => '0');
      imageYSizeRSig <= (others => '0');
    elsif (stateRSig = READIN and horizBlankIn = '0' and 
	   clkIn'event and clkIn = '0') then
      imageXSizeRSig <= addressRSig(3 downto 0);
      imageYSizeRSig <= addressRSig(7 downto 4);
    end if;
  end process; -- latchImageSize

  latchAddress: process(clkIn, resetLowIn)
  begin
    if (resetLowIn = '0') then
      addressRSig <= (others => '0');
      stateRSig <= IDLE;
      halfClkSig <= '0';
    elsif (clkIn'event and clkIn = '1') then
      addressRSig <= nextAddrSig;
      stateRSig <= nextStateSig;
      halfClkSig <= not halfClkSig;
    end if;
  end process; -- latchAddress

  drivers: process(writeEnable, addressRSig)
  begin
    if(writeEnable = TRUE) then
      vidWriteEnLowOut <= '0';
    else
      vidWriteEnLowOut <= '1';
    end if;
    vidAddrOut <= addressRSig;
  end process; -- drivers

end rtl; -- modInputDecoder
